# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
import warnings

import qbitcoin.generated.qbit_pb2 as qbit__pb2

GRPC_GENERATED_VERSION = '1.72.1'
GRPC_VERSION = grpc.__version__
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower
    _version_not_supported = first_version_is_lower(GRPC_VERSION, GRPC_GENERATED_VERSION)
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    raise RuntimeError(
        f'The grpc package installed is at version {GRPC_VERSION},'
        + f' but the generated code in qbit_pb2_grpc.py depends on'
        + f' grpcio>={GRPC_GENERATED_VERSION}.'
        + f' Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}'
        + f' or downgrade your generated code using grpcio-tools<={GRPC_VERSION}.'
    )


class PublicAPIStub(object):
    """//////////////////////////
    //////////////////////////
    //////////////////////////
    ////     API       ///////
    //////////////////////////
    //////////////////////////
    //////////////////////////

    This service describes the Public API used by clients (wallet/cli/etc)
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.GetNodeState = channel.unary_unary(
                '/qrl.PublicAPI/GetNodeState',
                request_serializer=qbit__pb2.GetNodeStateReq.SerializeToString,
                response_deserializer=qbit__pb2.GetNodeStateResp.FromString,
                _registered_method=True)
        self.GetKnownPeers = channel.unary_unary(
                '/qrl.PublicAPI/GetKnownPeers',
                request_serializer=qbit__pb2.GetKnownPeersReq.SerializeToString,
                response_deserializer=qbit__pb2.GetKnownPeersResp.FromString,
                _registered_method=True)
        self.GetPeersStat = channel.unary_unary(
                '/qrl.PublicAPI/GetPeersStat',
                request_serializer=qbit__pb2.GetPeersStatReq.SerializeToString,
                response_deserializer=qbit__pb2.GetPeersStatResp.FromString,
                _registered_method=True)
        self.GetStats = channel.unary_unary(
                '/qrl.PublicAPI/GetStats',
                request_serializer=qbit__pb2.GetStatsReq.SerializeToString,
                response_deserializer=qbit__pb2.GetStatsResp.FromString,
                _registered_method=True)
        self.GetAddressState = channel.unary_unary(
                '/qrl.PublicAPI/GetAddressState',
                request_serializer=qbit__pb2.GetAddressStateReq.SerializeToString,
                response_deserializer=qbit__pb2.GetAddressStateResp.FromString,
                _registered_method=True)
        self.GetOptimizedAddressState = channel.unary_unary(
                '/qrl.PublicAPI/GetOptimizedAddressState',
                request_serializer=qbit__pb2.GetAddressStateReq.SerializeToString,
                response_deserializer=qbit__pb2.GetOptimizedAddressStateResp.FromString,
                _registered_method=True)
        self.GetMultiSigAddressState = channel.unary_unary(
                '/qrl.PublicAPI/GetMultiSigAddressState',
                request_serializer=qbit__pb2.GetMultiSigAddressStateReq.SerializeToString,
                response_deserializer=qbit__pb2.GetMultiSigAddressStateResp.FromString,
                _registered_method=True)
        self.IsSlave = channel.unary_unary(
                '/qrl.PublicAPI/IsSlave',
                request_serializer=qbit__pb2.IsSlaveReq.SerializeToString,
                response_deserializer=qbit__pb2.IsSlaveResp.FromString,
                _registered_method=True)
        self.GetObject = channel.unary_unary(
                '/qrl.PublicAPI/GetObject',
                request_serializer=qbit__pb2.GetObjectReq.SerializeToString,
                response_deserializer=qbit__pb2.GetObjectResp.FromString,
                _registered_method=True)
        self.GetLatestData = channel.unary_unary(
                '/qrl.PublicAPI/GetLatestData',
                request_serializer=qbit__pb2.GetLatestDataReq.SerializeToString,
                response_deserializer=qbit__pb2.GetLatestDataResp.FromString,
                _registered_method=True)
        self.PushTransaction = channel.unary_unary(
                '/qrl.PublicAPI/PushTransaction',
                request_serializer=qbit__pb2.PushTransactionReq.SerializeToString,
                response_deserializer=qbit__pb2.PushTransactionResp.FromString,
                _registered_method=True)
        self.TransferCoins = channel.unary_unary(
                '/qrl.PublicAPI/TransferCoins',
                request_serializer=qbit__pb2.TransferCoinsReq.SerializeToString,
                response_deserializer=qbit__pb2.TransferCoinsResp.FromString,
                _registered_method=True)
        self.ParseAddress = channel.unary_unary(
                '/qrl.PublicAPI/ParseAddress',
                request_serializer=qbit__pb2.ParseAddressReq.SerializeToString,
                response_deserializer=qbit__pb2.ParseAddressResp.FromString,
                _registered_method=True)
        self.GetChainStats = channel.unary_unary(
                '/qrl.PublicAPI/GetChainStats',
                request_serializer=qbit__pb2.GetChainStatsReq.SerializeToString,
                response_deserializer=qbit__pb2.GetChainStatsResp.FromString,
                _registered_method=True)
        self.GetStakingStats = channel.unary_unary(
                '/qrl.PublicAPI/GetStakingStats',
                request_serializer=qbit__pb2.GetStakingStatsReq.SerializeToString,
                response_deserializer=qbit__pb2.GetStakingStatsResp.FromString,
                _registered_method=True)
        self.GetStakerList = channel.unary_unary(
                '/qrl.PublicAPI/GetStakerList',
                request_serializer=qbit__pb2.GetStakerListReq.SerializeToString,
                response_deserializer=qbit__pb2.GetStakerListResp.FromString,
                _registered_method=True)
        self.GetAddressFromPK = channel.unary_unary(
                '/qrl.PublicAPI/GetAddressFromPK',
                request_serializer=qbit__pb2.GetAddressFromPKReq.SerializeToString,
                response_deserializer=qbit__pb2.GetAddressFromPKResp.FromString,
                _registered_method=True)
        self.GetMultiSigCreateTxn = channel.unary_unary(
                '/qrl.PublicAPI/GetMultiSigCreateTxn',
                request_serializer=qbit__pb2.MultiSigCreateTxnReq.SerializeToString,
                response_deserializer=qbit__pb2.TransferCoinsResp.FromString,
                _registered_method=True)
        self.GetMultiSigSpendTxn = channel.unary_unary(
                '/qrl.PublicAPI/GetMultiSigSpendTxn',
                request_serializer=qbit__pb2.MultiSigSpendTxnReq.SerializeToString,
                response_deserializer=qbit__pb2.TransferCoinsResp.FromString,
                _registered_method=True)
        self.GetMultiSigVoteTxn = channel.unary_unary(
                '/qrl.PublicAPI/GetMultiSigVoteTxn',
                request_serializer=qbit__pb2.MultiSigVoteTxnReq.SerializeToString,
                response_deserializer=qbit__pb2.TransferCoinsResp.FromString,
                _registered_method=True)
        self.GetMessageTxn = channel.unary_unary(
                '/qrl.PublicAPI/GetMessageTxn',
                request_serializer=qbit__pb2.MessageTxnReq.SerializeToString,
                response_deserializer=qbit__pb2.TransferCoinsResp.FromString,
                _registered_method=True)
        self.GetTokenTxn = channel.unary_unary(
                '/qrl.PublicAPI/GetTokenTxn',
                request_serializer=qbit__pb2.TokenTxnReq.SerializeToString,
                response_deserializer=qbit__pb2.TransferCoinsResp.FromString,
                _registered_method=True)
        self.GetTransferTokenTxn = channel.unary_unary(
                '/qrl.PublicAPI/GetTransferTokenTxn',
                request_serializer=qbit__pb2.TransferTokenTxnReq.SerializeToString,
                response_deserializer=qbit__pb2.TransferCoinsResp.FromString,
                _registered_method=True)
        self.GetSlaveTxn = channel.unary_unary(
                '/qrl.PublicAPI/GetSlaveTxn',
                request_serializer=qbit__pb2.SlaveTxnReq.SerializeToString,
                response_deserializer=qbit__pb2.TransferCoinsResp.FromString,
                _registered_method=True)
        self.GetLatticeTxn = channel.unary_unary(
                '/qrl.PublicAPI/GetLatticeTxn',
                request_serializer=qbit__pb2.LatticeTxnReq.SerializeToString,
                response_deserializer=qbit__pb2.TransferCoinsResp.FromString,
                _registered_method=True)
        self.GetTransaction = channel.unary_unary(
                '/qrl.PublicAPI/GetTransaction',
                request_serializer=qbit__pb2.GetTransactionReq.SerializeToString,
                response_deserializer=qbit__pb2.GetTransactionResp.FromString,
                _registered_method=True)
        self.GetMiniTransactionsByAddress = channel.unary_unary(
                '/qrl.PublicAPI/GetMiniTransactionsByAddress',
                request_serializer=qbit__pb2.GetMiniTransactionsByAddressReq.SerializeToString,
                response_deserializer=qbit__pb2.GetMiniTransactionsByAddressResp.FromString,
                _registered_method=True)
        self.GetTransactionsByAddress = channel.unary_unary(
                '/qrl.PublicAPI/GetTransactionsByAddress',
                request_serializer=qbit__pb2.GetTransactionsByAddressReq.SerializeToString,
                response_deserializer=qbit__pb2.GetTransactionsByAddressResp.FromString,
                _registered_method=True)
        self.GetTokensByAddress = channel.unary_unary(
                '/qrl.PublicAPI/GetTokensByAddress',
                request_serializer=qbit__pb2.GetTransactionsByAddressReq.SerializeToString,
                response_deserializer=qbit__pb2.GetTokensByAddressResp.FromString,
                _registered_method=True)
        self.GetSlavesByAddress = channel.unary_unary(
                '/qrl.PublicAPI/GetSlavesByAddress',
                request_serializer=qbit__pb2.GetTransactionsByAddressReq.SerializeToString,
                response_deserializer=qbit__pb2.GetSlavesByAddressResp.FromString,
                _registered_method=True)
        self.GetLatticePKsByAddress = channel.unary_unary(
                '/qrl.PublicAPI/GetLatticePKsByAddress',
                request_serializer=qbit__pb2.GetTransactionsByAddressReq.SerializeToString,
                response_deserializer=qbit__pb2.GetLatticePKsByAddressResp.FromString,
                _registered_method=True)
        self.GetMultiSigAddressesByAddress = channel.unary_unary(
                '/qrl.PublicAPI/GetMultiSigAddressesByAddress',
                request_serializer=qbit__pb2.GetTransactionsByAddressReq.SerializeToString,
                response_deserializer=qbit__pb2.GetMultiSigAddressesByAddressResp.FromString,
                _registered_method=True)
        self.GetMultiSigSpendTxsByAddress = channel.unary_unary(
                '/qrl.PublicAPI/GetMultiSigSpendTxsByAddress',
                request_serializer=qbit__pb2.GetMultiSigSpendTxsByAddressReq.SerializeToString,
                response_deserializer=qbit__pb2.GetMultiSigSpendTxsByAddressResp.FromString,
                _registered_method=True)
        self.GetVoteStats = channel.unary_unary(
                '/qrl.PublicAPI/GetVoteStats',
                request_serializer=qbit__pb2.GetVoteStatsReq.SerializeToString,
                response_deserializer=qbit__pb2.GetVoteStatsResp.FromString,
                _registered_method=True)
        self.GetInboxMessagesByAddress = channel.unary_unary(
                '/qrl.PublicAPI/GetInboxMessagesByAddress',
                request_serializer=qbit__pb2.GetTransactionsByAddressReq.SerializeToString,
                response_deserializer=qbit__pb2.GetInboxMessagesByAddressResp.FromString,
                _registered_method=True)
        self.GetBalance = channel.unary_unary(
                '/qrl.PublicAPI/GetBalance',
                request_serializer=qbit__pb2.GetBalanceReq.SerializeToString,
                response_deserializer=qbit__pb2.GetBalanceResp.FromString,
                _registered_method=True)
        self.GetTotalBalance = channel.unary_unary(
                '/qrl.PublicAPI/GetTotalBalance',
                request_serializer=qbit__pb2.GetTotalBalanceReq.SerializeToString,
                response_deserializer=qbit__pb2.GetTotalBalanceResp.FromString,
                _registered_method=True)
        self.GetOTS = channel.unary_unary(
                '/qrl.PublicAPI/GetOTS',
                request_serializer=qbit__pb2.GetOTSReq.SerializeToString,
                response_deserializer=qbit__pb2.GetOTSResp.FromString,
                _registered_method=True)
        self.GetHeight = channel.unary_unary(
                '/qrl.PublicAPI/GetHeight',
                request_serializer=qbit__pb2.GetHeightReq.SerializeToString,
                response_deserializer=qbit__pb2.GetHeightResp.FromString,
                _registered_method=True)
        self.GetBlock = channel.unary_unary(
                '/qrl.PublicAPI/GetBlock',
                request_serializer=qbit__pb2.GetBlockReq.SerializeToString,
                response_deserializer=qbit__pb2.GetBlockResp.FromString,
                _registered_method=True)
        self.GetBlockByNumber = channel.unary_unary(
                '/qrl.PublicAPI/GetBlockByNumber',
                request_serializer=qbit__pb2.GetBlockByNumberReq.SerializeToString,
                response_deserializer=qbit__pb2.GetBlockByNumberResp.FromString,
                _registered_method=True)


class PublicAPIServicer(object):
    """//////////////////////////
    //////////////////////////
    //////////////////////////
    ////     API       ///////
    //////////////////////////
    //////////////////////////
    //////////////////////////

    This service describes the Public API used by clients (wallet/cli/etc)
    """

    def GetNodeState(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetKnownPeers(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetPeersStat(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetStats(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetAddressState(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetOptimizedAddressState(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetMultiSigAddressState(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def IsSlave(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetObject(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetLatestData(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def PushTransaction(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def TransferCoins(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ParseAddress(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetChainStats(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetStakingStats(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetStakerList(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetAddressFromPK(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetMultiSigCreateTxn(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetMultiSigSpendTxn(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetMultiSigVoteTxn(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetMessageTxn(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetTokenTxn(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetTransferTokenTxn(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetSlaveTxn(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetLatticeTxn(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetTransaction(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetMiniTransactionsByAddress(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetTransactionsByAddress(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetTokensByAddress(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetSlavesByAddress(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetLatticePKsByAddress(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetMultiSigAddressesByAddress(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetMultiSigSpendTxsByAddress(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetVoteStats(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetInboxMessagesByAddress(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetBalance(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetTotalBalance(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetOTS(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetHeight(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetBlock(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetBlockByNumber(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_PublicAPIServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'GetNodeState': grpc.unary_unary_rpc_method_handler(
                    servicer.GetNodeState,
                    request_deserializer=qbit__pb2.GetNodeStateReq.FromString,
                    response_serializer=qbit__pb2.GetNodeStateResp.SerializeToString,
            ),
            'GetKnownPeers': grpc.unary_unary_rpc_method_handler(
                    servicer.GetKnownPeers,
                    request_deserializer=qbit__pb2.GetKnownPeersReq.FromString,
                    response_serializer=qbit__pb2.GetKnownPeersResp.SerializeToString,
            ),
            'GetPeersStat': grpc.unary_unary_rpc_method_handler(
                    servicer.GetPeersStat,
                    request_deserializer=qbit__pb2.GetPeersStatReq.FromString,
                    response_serializer=qbit__pb2.GetPeersStatResp.SerializeToString,
            ),
            'GetStats': grpc.unary_unary_rpc_method_handler(
                    servicer.GetStats,
                    request_deserializer=qbit__pb2.GetStatsReq.FromString,
                    response_serializer=qbit__pb2.GetStatsResp.SerializeToString,
            ),
            'GetAddressState': grpc.unary_unary_rpc_method_handler(
                    servicer.GetAddressState,
                    request_deserializer=qbit__pb2.GetAddressStateReq.FromString,
                    response_serializer=qbit__pb2.GetAddressStateResp.SerializeToString,
            ),
            'GetOptimizedAddressState': grpc.unary_unary_rpc_method_handler(
                    servicer.GetOptimizedAddressState,
                    request_deserializer=qbit__pb2.GetAddressStateReq.FromString,
                    response_serializer=qbit__pb2.GetOptimizedAddressStateResp.SerializeToString,
            ),
            'GetMultiSigAddressState': grpc.unary_unary_rpc_method_handler(
                    servicer.GetMultiSigAddressState,
                    request_deserializer=qbit__pb2.GetMultiSigAddressStateReq.FromString,
                    response_serializer=qbit__pb2.GetMultiSigAddressStateResp.SerializeToString,
            ),
            'IsSlave': grpc.unary_unary_rpc_method_handler(
                    servicer.IsSlave,
                    request_deserializer=qbit__pb2.IsSlaveReq.FromString,
                    response_serializer=qbit__pb2.IsSlaveResp.SerializeToString,
            ),
            'GetObject': grpc.unary_unary_rpc_method_handler(
                    servicer.GetObject,
                    request_deserializer=qbit__pb2.GetObjectReq.FromString,
                    response_serializer=qbit__pb2.GetObjectResp.SerializeToString,
            ),
            'GetLatestData': grpc.unary_unary_rpc_method_handler(
                    servicer.GetLatestData,
                    request_deserializer=qbit__pb2.GetLatestDataReq.FromString,
                    response_serializer=qbit__pb2.GetLatestDataResp.SerializeToString,
            ),
            'PushTransaction': grpc.unary_unary_rpc_method_handler(
                    servicer.PushTransaction,
                    request_deserializer=qbit__pb2.PushTransactionReq.FromString,
                    response_serializer=qbit__pb2.PushTransactionResp.SerializeToString,
            ),
            'TransferCoins': grpc.unary_unary_rpc_method_handler(
                    servicer.TransferCoins,
                    request_deserializer=qbit__pb2.TransferCoinsReq.FromString,
                    response_serializer=qbit__pb2.TransferCoinsResp.SerializeToString,
            ),
            'ParseAddress': grpc.unary_unary_rpc_method_handler(
                    servicer.ParseAddress,
                    request_deserializer=qbit__pb2.ParseAddressReq.FromString,
                    response_serializer=qbit__pb2.ParseAddressResp.SerializeToString,
            ),
            'GetChainStats': grpc.unary_unary_rpc_method_handler(
                    servicer.GetChainStats,
                    request_deserializer=qbit__pb2.GetChainStatsReq.FromString,
                    response_serializer=qbit__pb2.GetChainStatsResp.SerializeToString,
            ),
            'GetStakingStats': grpc.unary_unary_rpc_method_handler(
                    servicer.GetStakingStats,
                    request_deserializer=qbit__pb2.GetStakingStatsReq.FromString,
                    response_serializer=qbit__pb2.GetStakingStatsResp.SerializeToString,
            ),
            'GetStakerList': grpc.unary_unary_rpc_method_handler(
                    servicer.GetStakerList,
                    request_deserializer=qbit__pb2.GetStakerListReq.FromString,
                    response_serializer=qbit__pb2.GetStakerListResp.SerializeToString,
            ),
            'GetAddressFromPK': grpc.unary_unary_rpc_method_handler(
                    servicer.GetAddressFromPK,
                    request_deserializer=qbit__pb2.GetAddressFromPKReq.FromString,
                    response_serializer=qbit__pb2.GetAddressFromPKResp.SerializeToString,
            ),
            'GetMultiSigCreateTxn': grpc.unary_unary_rpc_method_handler(
                    servicer.GetMultiSigCreateTxn,
                    request_deserializer=qbit__pb2.MultiSigCreateTxnReq.FromString,
                    response_serializer=qbit__pb2.TransferCoinsResp.SerializeToString,
            ),
            'GetMultiSigSpendTxn': grpc.unary_unary_rpc_method_handler(
                    servicer.GetMultiSigSpendTxn,
                    request_deserializer=qbit__pb2.MultiSigSpendTxnReq.FromString,
                    response_serializer=qbit__pb2.TransferCoinsResp.SerializeToString,
            ),
            'GetMultiSigVoteTxn': grpc.unary_unary_rpc_method_handler(
                    servicer.GetMultiSigVoteTxn,
                    request_deserializer=qbit__pb2.MultiSigVoteTxnReq.FromString,
                    response_serializer=qbit__pb2.TransferCoinsResp.SerializeToString,
            ),
            'GetMessageTxn': grpc.unary_unary_rpc_method_handler(
                    servicer.GetMessageTxn,
                    request_deserializer=qbit__pb2.MessageTxnReq.FromString,
                    response_serializer=qbit__pb2.TransferCoinsResp.SerializeToString,
            ),
            'GetTokenTxn': grpc.unary_unary_rpc_method_handler(
                    servicer.GetTokenTxn,
                    request_deserializer=qbit__pb2.TokenTxnReq.FromString,
                    response_serializer=qbit__pb2.TransferCoinsResp.SerializeToString,
            ),
            'GetTransferTokenTxn': grpc.unary_unary_rpc_method_handler(
                    servicer.GetTransferTokenTxn,
                    request_deserializer=qbit__pb2.TransferTokenTxnReq.FromString,
                    response_serializer=qbit__pb2.TransferCoinsResp.SerializeToString,
            ),
            'GetSlaveTxn': grpc.unary_unary_rpc_method_handler(
                    servicer.GetSlaveTxn,
                    request_deserializer=qbit__pb2.SlaveTxnReq.FromString,
                    response_serializer=qbit__pb2.TransferCoinsResp.SerializeToString,
            ),
            'GetLatticeTxn': grpc.unary_unary_rpc_method_handler(
                    servicer.GetLatticeTxn,
                    request_deserializer=qbit__pb2.LatticeTxnReq.FromString,
                    response_serializer=qbit__pb2.TransferCoinsResp.SerializeToString,
            ),
            'GetTransaction': grpc.unary_unary_rpc_method_handler(
                    servicer.GetTransaction,
                    request_deserializer=qbit__pb2.GetTransactionReq.FromString,
                    response_serializer=qbit__pb2.GetTransactionResp.SerializeToString,
            ),
            'GetMiniTransactionsByAddress': grpc.unary_unary_rpc_method_handler(
                    servicer.GetMiniTransactionsByAddress,
                    request_deserializer=qbit__pb2.GetMiniTransactionsByAddressReq.FromString,
                    response_serializer=qbit__pb2.GetMiniTransactionsByAddressResp.SerializeToString,
            ),
            'GetTransactionsByAddress': grpc.unary_unary_rpc_method_handler(
                    servicer.GetTransactionsByAddress,
                    request_deserializer=qbit__pb2.GetTransactionsByAddressReq.FromString,
                    response_serializer=qbit__pb2.GetTransactionsByAddressResp.SerializeToString,
            ),
            'GetTokensByAddress': grpc.unary_unary_rpc_method_handler(
                    servicer.GetTokensByAddress,
                    request_deserializer=qbit__pb2.GetTransactionsByAddressReq.FromString,
                    response_serializer=qbit__pb2.GetTokensByAddressResp.SerializeToString,
            ),
            'GetSlavesByAddress': grpc.unary_unary_rpc_method_handler(
                    servicer.GetSlavesByAddress,
                    request_deserializer=qbit__pb2.GetTransactionsByAddressReq.FromString,
                    response_serializer=qbit__pb2.GetSlavesByAddressResp.SerializeToString,
            ),
            'GetLatticePKsByAddress': grpc.unary_unary_rpc_method_handler(
                    servicer.GetLatticePKsByAddress,
                    request_deserializer=qbit__pb2.GetTransactionsByAddressReq.FromString,
                    response_serializer=qbit__pb2.GetLatticePKsByAddressResp.SerializeToString,
            ),
            'GetMultiSigAddressesByAddress': grpc.unary_unary_rpc_method_handler(
                    servicer.GetMultiSigAddressesByAddress,
                    request_deserializer=qbit__pb2.GetTransactionsByAddressReq.FromString,
                    response_serializer=qbit__pb2.GetMultiSigAddressesByAddressResp.SerializeToString,
            ),
            'GetMultiSigSpendTxsByAddress': grpc.unary_unary_rpc_method_handler(
                    servicer.GetMultiSigSpendTxsByAddress,
                    request_deserializer=qbit__pb2.GetMultiSigSpendTxsByAddressReq.FromString,
                    response_serializer=qbit__pb2.GetMultiSigSpendTxsByAddressResp.SerializeToString,
            ),
            'GetVoteStats': grpc.unary_unary_rpc_method_handler(
                    servicer.GetVoteStats,
                    request_deserializer=qbit__pb2.GetVoteStatsReq.FromString,
                    response_serializer=qbit__pb2.GetVoteStatsResp.SerializeToString,
            ),
            'GetInboxMessagesByAddress': grpc.unary_unary_rpc_method_handler(
                    servicer.GetInboxMessagesByAddress,
                    request_deserializer=qbit__pb2.GetTransactionsByAddressReq.FromString,
                    response_serializer=qbit__pb2.GetInboxMessagesByAddressResp.SerializeToString,
            ),
            'GetBalance': grpc.unary_unary_rpc_method_handler(
                    servicer.GetBalance,
                    request_deserializer=qbit__pb2.GetBalanceReq.FromString,
                    response_serializer=qbit__pb2.GetBalanceResp.SerializeToString,
            ),
            'GetTotalBalance': grpc.unary_unary_rpc_method_handler(
                    servicer.GetTotalBalance,
                    request_deserializer=qbit__pb2.GetTotalBalanceReq.FromString,
                    response_serializer=qbit__pb2.GetTotalBalanceResp.SerializeToString,
            ),
            'GetOTS': grpc.unary_unary_rpc_method_handler(
                    servicer.GetOTS,
                    request_deserializer=qbit__pb2.GetOTSReq.FromString,
                    response_serializer=qbit__pb2.GetOTSResp.SerializeToString,
            ),
            'GetHeight': grpc.unary_unary_rpc_method_handler(
                    servicer.GetHeight,
                    request_deserializer=qbit__pb2.GetHeightReq.FromString,
                    response_serializer=qbit__pb2.GetHeightResp.SerializeToString,
            ),
            'GetBlock': grpc.unary_unary_rpc_method_handler(
                    servicer.GetBlock,
                    request_deserializer=qbit__pb2.GetBlockReq.FromString,
                    response_serializer=qbit__pb2.GetBlockResp.SerializeToString,
            ),
            'GetBlockByNumber': grpc.unary_unary_rpc_method_handler(
                    servicer.GetBlockByNumber,
                    request_deserializer=qbit__pb2.GetBlockByNumberReq.FromString,
                    response_serializer=qbit__pb2.GetBlockByNumberResp.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'qrl.PublicAPI', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('qrl.PublicAPI', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class PublicAPI(object):
    """//////////////////////////
    //////////////////////////
    //////////////////////////
    ////     API       ///////
    //////////////////////////
    //////////////////////////
    //////////////////////////

    This service describes the Public API used by clients (wallet/cli/etc)
    """

    @staticmethod
    def GetNodeState(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qrl.PublicAPI/GetNodeState',
            qbit__pb2.GetNodeStateReq.SerializeToString,
            qbit__pb2.GetNodeStateResp.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetKnownPeers(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qrl.PublicAPI/GetKnownPeers',
            qbit__pb2.GetKnownPeersReq.SerializeToString,
            qbit__pb2.GetKnownPeersResp.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetPeersStat(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qrl.PublicAPI/GetPeersStat',
            qbit__pb2.GetPeersStatReq.SerializeToString,
            qbit__pb2.GetPeersStatResp.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetStats(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qrl.PublicAPI/GetStats',
            qbit__pb2.GetStatsReq.SerializeToString,
            qbit__pb2.GetStatsResp.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetAddressState(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qrl.PublicAPI/GetAddressState',
            qbit__pb2.GetAddressStateReq.SerializeToString,
            qbit__pb2.GetAddressStateResp.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetOptimizedAddressState(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qrl.PublicAPI/GetOptimizedAddressState',
            qbit__pb2.GetAddressStateReq.SerializeToString,
            qbit__pb2.GetOptimizedAddressStateResp.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetMultiSigAddressState(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qrl.PublicAPI/GetMultiSigAddressState',
            qbit__pb2.GetMultiSigAddressStateReq.SerializeToString,
            qbit__pb2.GetMultiSigAddressStateResp.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def IsSlave(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qrl.PublicAPI/IsSlave',
            qbit__pb2.IsSlaveReq.SerializeToString,
            qbit__pb2.IsSlaveResp.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetObject(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qrl.PublicAPI/GetObject',
            qbit__pb2.GetObjectReq.SerializeToString,
            qbit__pb2.GetObjectResp.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetLatestData(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qrl.PublicAPI/GetLatestData',
            qbit__pb2.GetLatestDataReq.SerializeToString,
            qbit__pb2.GetLatestDataResp.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def PushTransaction(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qrl.PublicAPI/PushTransaction',
            qbit__pb2.PushTransactionReq.SerializeToString,
            qbit__pb2.PushTransactionResp.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def TransferCoins(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qrl.PublicAPI/TransferCoins',
            qbit__pb2.TransferCoinsReq.SerializeToString,
            qbit__pb2.TransferCoinsResp.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def ParseAddress(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qrl.PublicAPI/ParseAddress',
            qbit__pb2.ParseAddressReq.SerializeToString,
            qbit__pb2.ParseAddressResp.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetChainStats(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qrl.PublicAPI/GetChainStats',
            qbit__pb2.GetChainStatsReq.SerializeToString,
            qbit__pb2.GetChainStatsResp.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetStakingStats(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qrl.PublicAPI/GetStakingStats',
            qbit__pb2.GetStakingStatsReq.SerializeToString,
            qbit__pb2.GetStakingStatsResp.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetStakerList(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qrl.PublicAPI/GetStakerList',
            qbit__pb2.GetStakerListReq.SerializeToString,
            qbit__pb2.GetStakerListResp.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetAddressFromPK(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qrl.PublicAPI/GetAddressFromPK',
            qbit__pb2.GetAddressFromPKReq.SerializeToString,
            qbit__pb2.GetAddressFromPKResp.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetMultiSigCreateTxn(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qrl.PublicAPI/GetMultiSigCreateTxn',
            qbit__pb2.MultiSigCreateTxnReq.SerializeToString,
            qbit__pb2.TransferCoinsResp.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetMultiSigSpendTxn(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qrl.PublicAPI/GetMultiSigSpendTxn',
            qbit__pb2.MultiSigSpendTxnReq.SerializeToString,
            qbit__pb2.TransferCoinsResp.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetMultiSigVoteTxn(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qrl.PublicAPI/GetMultiSigVoteTxn',
            qbit__pb2.MultiSigVoteTxnReq.SerializeToString,
            qbit__pb2.TransferCoinsResp.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetMessageTxn(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qrl.PublicAPI/GetMessageTxn',
            qbit__pb2.MessageTxnReq.SerializeToString,
            qbit__pb2.TransferCoinsResp.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetTokenTxn(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qrl.PublicAPI/GetTokenTxn',
            qbit__pb2.TokenTxnReq.SerializeToString,
            qbit__pb2.TransferCoinsResp.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetTransferTokenTxn(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qrl.PublicAPI/GetTransferTokenTxn',
            qbit__pb2.TransferTokenTxnReq.SerializeToString,
            qbit__pb2.TransferCoinsResp.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetSlaveTxn(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qrl.PublicAPI/GetSlaveTxn',
            qbit__pb2.SlaveTxnReq.SerializeToString,
            qbit__pb2.TransferCoinsResp.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetLatticeTxn(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qrl.PublicAPI/GetLatticeTxn',
            qbit__pb2.LatticeTxnReq.SerializeToString,
            qbit__pb2.TransferCoinsResp.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetTransaction(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qrl.PublicAPI/GetTransaction',
            qbit__pb2.GetTransactionReq.SerializeToString,
            qbit__pb2.GetTransactionResp.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetMiniTransactionsByAddress(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qrl.PublicAPI/GetMiniTransactionsByAddress',
            qbit__pb2.GetMiniTransactionsByAddressReq.SerializeToString,
            qbit__pb2.GetMiniTransactionsByAddressResp.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetTransactionsByAddress(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qrl.PublicAPI/GetTransactionsByAddress',
            qbit__pb2.GetTransactionsByAddressReq.SerializeToString,
            qbit__pb2.GetTransactionsByAddressResp.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetTokensByAddress(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qrl.PublicAPI/GetTokensByAddress',
            qbit__pb2.GetTransactionsByAddressReq.SerializeToString,
            qbit__pb2.GetTokensByAddressResp.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetSlavesByAddress(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qrl.PublicAPI/GetSlavesByAddress',
            qbit__pb2.GetTransactionsByAddressReq.SerializeToString,
            qbit__pb2.GetSlavesByAddressResp.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetLatticePKsByAddress(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qrl.PublicAPI/GetLatticePKsByAddress',
            qbit__pb2.GetTransactionsByAddressReq.SerializeToString,
            qbit__pb2.GetLatticePKsByAddressResp.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetMultiSigAddressesByAddress(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qrl.PublicAPI/GetMultiSigAddressesByAddress',
            qbit__pb2.GetTransactionsByAddressReq.SerializeToString,
            qbit__pb2.GetMultiSigAddressesByAddressResp.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetMultiSigSpendTxsByAddress(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qrl.PublicAPI/GetMultiSigSpendTxsByAddress',
            qbit__pb2.GetMultiSigSpendTxsByAddressReq.SerializeToString,
            qbit__pb2.GetMultiSigSpendTxsByAddressResp.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetVoteStats(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qrl.PublicAPI/GetVoteStats',
            qbit__pb2.GetVoteStatsReq.SerializeToString,
            qbit__pb2.GetVoteStatsResp.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetInboxMessagesByAddress(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qrl.PublicAPI/GetInboxMessagesByAddress',
            qbit__pb2.GetTransactionsByAddressReq.SerializeToString,
            qbit__pb2.GetInboxMessagesByAddressResp.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetBalance(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qrl.PublicAPI/GetBalance',
            qbit__pb2.GetBalanceReq.SerializeToString,
            qbit__pb2.GetBalanceResp.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetTotalBalance(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qrl.PublicAPI/GetTotalBalance',
            qbit__pb2.GetTotalBalanceReq.SerializeToString,
            qbit__pb2.GetTotalBalanceResp.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetOTS(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qrl.PublicAPI/GetOTS',
            qbit__pb2.GetOTSReq.SerializeToString,
            qbit__pb2.GetOTSResp.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetHeight(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qrl.PublicAPI/GetHeight',
            qbit__pb2.GetHeightReq.SerializeToString,
            qbit__pb2.GetHeightResp.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetBlock(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qrl.PublicAPI/GetBlock',
            qbit__pb2.GetBlockReq.SerializeToString,
            qbit__pb2.GetBlockResp.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetBlockByNumber(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qrl.PublicAPI/GetBlockByNumber',
            qbit__pb2.GetBlockByNumberReq.SerializeToString,
            qbit__pb2.GetBlockByNumberResp.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)


class AdminAPIStub(object):
    """This is a place holder for testing/instrumentation APIs
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """


class AdminAPIServicer(object):
    """This is a place holder for testing/instrumentation APIs
    """


def add_AdminAPIServicer_to_server(servicer, server):
    rpc_method_handlers = {
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'qrl.AdminAPI', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('qrl.AdminAPI', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class AdminAPI(object):
    """This is a place holder for testing/instrumentation APIs
    """



#!/usr/bin/env python3
# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.

"""
Qbitcoin Donor Wallet Management CLI Tool

This tool provides command-line interface for managing donor wallets:
- Create encrypted donor wallet
- Load donor wallet
- Check donor wallet status
- Verify donor wallet password
"""

import argparse
import sys
import getpass
import os
from pyqrllib.pyqrllib import hstr2bin, bin2hstr

from qbitcoin.core.config import config
from qbitcoin.core.DonorWalletManager import DonorWalletManager


def parse_arguments():
    parser = argparse.ArgumentParser(description='Qbitcoin Donor Wallet Management Tool')
    
    subparsers = parser.add_subparsers(dest='command', help='Available commands')
    
    # Create donor wallet command
    create_parser = subparsers.add_parser('create', help='Create encrypted donor wallet')
    create_parser.add_argument('address', help='Donor wallet address (with Q prefix)')
    create_parser.add_argument('private_key', help='Private key in hex format')
    
    # Load donor wallet command  
    load_parser = subparsers.add_parser('load', help='Load and verify donor wallet')
    
    # Status command
    status_parser = subparsers.add_parser('status', help='Show donor wallet status')
    
    # Verify password command
    verify_parser = subparsers.add_parser('verify', help='Verify donor wallet password')
    
    # Path command
    path_parser = subparsers.add_parser('path', help='Show donor wallet file path')
    
    # Common arguments
    parser.add_argument('--qrldir', '-d', dest='qrl_dir', default=config.user.qrl_dir,
                        help="Use a different directory for node data/configuration")
    parser.add_argument('--network-type', dest='network_type', choices=['mainnet', 'testnet'],
                        default='mainnet', required=False, help="Network type")
    
    return parser.parse_args()


def validate_address(address: str) -> bool:
    """Validate Qbitcoin address format"""
    if not address.startswith('Q'):
        print("❌ Address must start with 'Q'")
        return False
    
    if len(address) != 51:  # Q + 50 hex chars
        print("❌ Invalid address length")
        return False
    
    try:
        bytes(hstr2bin(address[1:]))
        return True
    except Exception as e:
        print(f"❌ Invalid address format: {e}")
        return False


def validate_private_key(private_key: str) -> bool:
    """Validate private key format"""
    if len(private_key) < 64:  # Minimum length for private key
        print("❌ Private key too short")
        return False
    
    try:
        bytes(hstr2bin(private_key))
        return True
    except Exception as e:
        print(f"❌ Invalid private key format: {e}")
        return False


def get_password(prompt: str) -> str:
    """Get password from user with confirmation"""
    while True:
        password = getpass.getpass(prompt)
        if len(password) < 8:
            print("❌ Password must be at least 8 characters long")
            continue
        return password


def get_password_with_confirmation(prompt: str) -> str:
    """Get password from user with confirmation"""
    while True:
        password = get_password(prompt)
        confirm_password = getpass.getpass("Confirm password: ")
        
        if password == confirm_password:
            return password
        else:
            print("❌ Passwords do not match. Please try again.")


def cmd_create_donor_wallet(args):
    """Create encrypted donor wallet"""
    try:
        # Validate inputs
        if not validate_address(args.address):
            return False
            
        if not validate_private_key(args.private_key):
            return False
        
        # Get password
        password = get_password_with_confirmation("Enter password for donor wallet: ")
        
        # Create donor wallet manager
        donor_manager = DonorWalletManager()
        
        # Check if wallet already exists
        if donor_manager.wallet_exists():
            overwrite = input("⚠️  Donor wallet already exists. Overwrite? (y/N): ")
            if overwrite.lower() != 'y':
                print("❌ Operation cancelled")
                return False
        
        # Create wallet
        if donor_manager.create_donor_wallet(args.address, args.private_key, password):
            print(f"✅ Donor wallet created successfully!")
            print(f"📍 Wallet file: {donor_manager.get_wallet_path()}")
            print(f"🏛️  Address: {args.address}")
            print("⚠️  Keep your password safe - it cannot be recovered!")
            return True
        else:
            print("❌ Failed to create donor wallet")
            return False
            
    except Exception as e:
        print(f"❌ Error creating donor wallet: {e}")
        return False


def cmd_load_donor_wallet(args):
    """Load and verify donor wallet"""
    try:
        donor_manager = DonorWalletManager()
        
        if not donor_manager.wallet_exists():
            print("❌ No donor wallet found")
            print(f"📍 Expected location: {donor_manager.get_wallet_path()}")
            return False
        
        # Get password
        password = get_password("Enter donor wallet password: ")
        
        # Load wallet
        wallet_data = donor_manager.load_donor_wallet(password)
        
        if wallet_data:
            print("✅ Donor wallet loaded successfully!")
            print(f"🏛️  Address: {wallet_data['address']}")
            print(f"📅 Created: {wallet_data.get('created_at', 'Unknown')}")
            print(f"🔢 Version: {wallet_data.get('version', 'Unknown')}")
            return True
        else:
            print("❌ Failed to load donor wallet (wrong password?)")
            return False
            
    except Exception as e:
        print(f"❌ Error loading donor wallet: {e}")
        return False


def cmd_show_status(args):
    """Show donor wallet status"""
    try:
        donor_manager = DonorWalletManager()
        
        print("🏛️  DONOR WALLET STATUS")
        print("=" * 40)
        
        if donor_manager.wallet_exists():
            print(f"📍 Wallet File: {donor_manager.get_wallet_path()}")
            print("📁 File Status: ✅ Exists")
            
            # Check if it's loaded
            if donor_manager.is_wallet_loaded():
                wallet_info = donor_manager.get_wallet_info()
                print("🔓 Load Status: ✅ Loaded")
                print(f"🏛️  Address: {wallet_info['address']}")
                print(f"📅 Created: {wallet_info.get('created_at', 'Unknown')}")
            else:
                print("🔓 Load Status: ❌ Not loaded")
        else:
            print(f"📍 Wallet File: {donor_manager.get_wallet_path()}")
            print("📁 File Status: ❌ Not found")
            print("🔓 Load Status: ❌ No wallet")
        
        return True
        
    except Exception as e:
        print(f"❌ Error getting donor wallet status: {e}")
        return False


def cmd_verify_password(args):
    """Verify donor wallet password"""
    try:
        donor_manager = DonorWalletManager()
        
        if not donor_manager.wallet_exists():
            print("❌ No donor wallet found")
            return False
        
        # Get password
        password = get_password("Enter password to verify: ")
        
        # Verify password
        if donor_manager.verify_password(password):
            print("✅ Password is correct!")
            return True
        else:
            print("❌ Password is incorrect!")
            return False
            
    except Exception as e:
        print(f"❌ Error verifying password: {e}")
        return False


def cmd_show_path(args):
    """Show donor wallet file path"""
    try:
        donor_manager = DonorWalletManager()
        
        print("📍 DONOR WALLET PATH")
        print("=" * 40)
        print(f"File Path: {donor_manager.get_wallet_path()}")
        print(f"Directory: {os.path.dirname(donor_manager.get_wallet_path())}")
        print(f"Exists: {'✅ Yes' if donor_manager.wallet_exists() else '❌ No'}")
        
        return True
        
    except Exception as e:
        print(f"❌ Error getting path: {e}")
        return False


def main():
    args = parse_arguments()
    
    # Set up configuration
    if args.network_type == 'testnet':
        config.dev.hard_fork_heights = list(config.dev.testnet_hard_fork_heights)
        config.dev.hard_fork_node_disconnect_delay = list(config.dev.testnet_hard_fork_node_disconnect_delay)
    
    config.user.qrl_dir = args.qrl_dir
    config.create_path(config.user.qrl_dir)
    
    # Execute command
    success = False
    
    if args.command == 'create':
        success = cmd_create_donor_wallet(args)
    elif args.command == 'load':
        success = cmd_load_donor_wallet(args)
    elif args.command == 'status':
        success = cmd_show_status(args)
    elif args.command == 'verify':
        success = cmd_verify_password(args)
    elif args.command == 'path':
        success = cmd_show_path(args)
    else:
        print("❌ No command specified. Use --help for usage information.")
        return 1
    
    return 0 if success else 1


if __name__ == '__main__':
    sys.exit(main())



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: qbitbase.proto
# Protobuf Python Version: 6.30.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    30,
    0,
    '',
    'qbitbase.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0eqbitbase.proto\x12\x03qrl\"\x10\n\x0eGetNodeInfoReq\"5\n\x0fGetNodeInfoResp\x12\x0f\n\x07version\x18\x01 \x01(\t\x12\x11\n\tgrpcProto\x18\x02 \x01(\t2B\n\x04\x42\x61se\x12:\n\x0bGetNodeInfo\x12\x13.qrl.GetNodeInfoReq\x1a\x14.qrl.GetNodeInfoResp\"\x00\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'qbitbase_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_GETNODEINFOREQ']._serialized_start=23
  _globals['_GETNODEINFOREQ']._serialized_end=39
  _globals['_GETNODEINFORESP']._serialized_start=41
  _globals['_GETNODEINFORESP']._serialized_end=94
  _globals['_BASE']._serialized_start=96
  _globals['_BASE']._serialized_end=162
# @@protoc_insertion_point(module_scope)



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.

from abc import ABCMeta, abstractmethod
from math import log, floor

from google.protobuf.json_format import MessageToJson, Parse
from pyqrllib.pyqrllib import bin2hstr

from qbitcoin.core import config
from qbitcoin.core.State import State
from qbitcoin.core.OptimizedAddressState import OptimizedAddressState
from qbitcoin.core.StateContainer import StateContainer
from qbitcoin.core.misc import logger
from qbitcoin.core.txs import build_tx as main_build_tx
from qbitcoin.core.FalconHelper import falcon_pk_to_address
from qbitcoin.core.FalconHelper import falcon_pk_to_address, is_falcon_public_key
from qbitcoin.crypto.misc import sha256
from qbitcoin.crypto.falcon import FalconSignature
from qbitcoin.generated import qbit_pb2

CODEMAP = {
    'transfer': 1,
    'coinbase': 2,
    'latticePK': 3,
    'message': 4,
    'token': 5,
    'transfer_token': 6,
    'slave': 7,
    'multi_sig_create': 8,
    'multi_sig_spend': 9,
    'multi_sig_vote': 10,
}


class Transaction(object, metaclass=ABCMeta):
    """
    Abstract Base class to be derived by all other transactions
    """

    def __init__(self, protobuf_transaction=None):
        self._data = protobuf_transaction  # This object cointains persistable data
        if protobuf_transaction is None:
            self._data = qbit_pb2.Transaction()

    def __lt__(self, tx):
        if self.fee < tx.fee:
            return True
        return False

    def __gt__(self, tx):
        if self.fee > tx.fee:
            return True
        return False

    @property
    def size(self):
        return self._data.ByteSize()

    @property
    def pbdata(self):
        """
        Returns a protobuf object that contains persistable data representing this object
        :return: A protobuf Transaction object
        :rtype: qbit_pb2.Transaction
        """
        return self._data

    @property
    def type(self):
        return self._data.WhichOneof('transactionType')

    @property
    def fee(self):
        return self._data.fee

    @property
    def nonce(self):
        return self._data.nonce

    @property
    def master_addr(self):
        return self._data.master_addr

    @property
    def addr_from(self):
        if self.master_addr:
            return self.master_addr

        return falcon_pk_to_address(self.PK)

    @property
    def ots_key(self):
        # With Falcon, we don't have OTS keys
        # Instead, we identify transactions by their public key
        return None

    @staticmethod
    def calc_allowed_decimals(value):
        if not isinstance(value, int):
            raise ValueError('value should be of integer type')
        if value == 0:
            return 19

        # floor value could be negative, so return 0 when the floor value is negative
        return max(floor(19 - log(value, 10)), 0)

    @property
    def PK(self):
        return self._data.public_key

    @property
    def signature(self):
        return self._data.signature

    @staticmethod
    def from_pbdata(pbdata: qbit_pb2.Transaction):
        return main_build_tx(pbdata.WhichOneof('transactionType'), pbdata)

    @staticmethod
    def from_json(json_data):
        pbdata = qbit_pb2.Transaction()
        Parse(json_data, pbdata)
        return Transaction.from_pbdata(pbdata)

    @staticmethod
    def get_slave(tx):
        addr_from_pk = falcon_pk_to_address(tx.PK)
        if addr_from_pk != tx.addr_from:
            return addr_from_pk
        return None

    @property
    def txhash(self) -> bytes:
        return self._data.transaction_hash

    def update_txhash(self):
        self._data.transaction_hash = self.generate_txhash()

    def generate_txhash(self):
        return sha256(
            self.get_data_hash() +
            self.signature +
            self.PK
        )

    def get_data_bytes(self) -> bytes:
        """
        This method returns the essential bytes that represent the transaction and will be later signed
        :return:
        """
        raise NotImplementedError

    def get_data_hash(self) -> bytes:
        """
        This method returns the hashes of the transaction data.
        :return:
        """
        return sha256(self.get_data_bytes())

    def sign(self, object_with_sign_method):
        self._data.signature = object_with_sign_method.sign(self.get_data_hash())
        self.update_txhash()

    def set_affected_address(self, addresses_set: set):
        addresses_set.add(self.addr_from)
        addresses_set.add(falcon_pk_to_address(self.PK))

    @abstractmethod
    def _validate_custom(self) -> bool:
        """
        This is an extension point for derived classes validation
        If derived classes need additional field validation they should override this member
        """
        raise NotImplementedError

    @abstractmethod
    def _validate_extended(self, state_container: StateContainer) -> bool:
        raise NotImplementedError

    @abstractmethod
    def apply(self,
              state: State,
              state_container: StateContainer) -> bool:
        raise NotImplementedError

    @abstractmethod
    def revert(self,
               state: State,
               state_container: StateContainer) -> bool:
        raise NotImplementedError

    def validate_transaction_pool(self, transaction_pool):
        for tx_set in transaction_pool:
            txn = tx_set[1].transaction
            if txn.txhash == self.txhash:
                continue

            if self.PK != txn.PK:
                continue

            # Falcon public key reuse validation removed - allowing multiple transactions with same public key

        return True

    def validate(self, verify_signature=True) -> bool:
        """
        This method calls validate_or_raise, logs any failure and returns True or False accordingly
        The main purpose is to avoid exceptions and accommodate legacy code
        :return: True is the transaction is valid
        :rtype: bool
        """
        try:
            self.validate_or_raise(verify_signature)
        except ValueError as e:
            logger.info('[%s] failed validate_tx', bin2hstr(self.txhash))
            logger.warning(str(e))
            return False
        except Exception as e:
            logger.exception(e)
            return False
        return True

    def validate_all(self, state_container: StateContainer, check_nonce=True) -> bool:
        # Check banned addresses only after hard fork height (index 0 = banned address enforcement)
        current_block_number = state_container.block_number
        banned_address_hard_fork_height = state_container.current_dev_config.hard_fork_heights[0]
        
        if current_block_number >= banned_address_hard_fork_height:
            for banned_address in state_container.current_dev_config.banned_address:
                tx_type = self.pbdata.WhichOneof('transactionType')
                addr_from_pk = None
                if tx_type != 'coinbase':
                    addr_from_pk = falcon_pk_to_address(self.PK)

                if addr_from_pk == banned_address or self.master_addr == banned_address:
                    logger.warning(f"Banned Qbitcoin Address found in master_addr or pk at block {current_block_number} (banned since block {banned_address_hard_fork_height})")
                    return False
                if tx_type == 'coinbase':
                    if self.pbdata.coinbase.addr_to == banned_address:
                        logger.warning(f"Banned Qbitcoin Address found in coinbase.addr_to at block {current_block_number}")
                        return False
                elif tx_type == 'message':
                    if self.pbdata.message.addr_to == banned_address:
                        logger.warning(f"Banned Qbitcoin Address found in message.addr_to at block {current_block_number}")
                        return False
                elif tx_type == 'transfer':
                    for addr_to in self.pbdata.transfer.addrs_to:
                        if banned_address == addr_to:
                            logger.warning(f"Banned Qbitcoin Address found in transfer.addr_to at block {current_block_number}")
                            return False
                elif tx_type == 'token':
                    # Token transactions don't have recipient addresses to check
                    # Only sender validation (already done above with addr_from_pk)
                    pass
                elif tx_type == 'transfer_token':
                    for addr_to in self.pbdata.transfer_token.addrs_to:
                        if banned_address == addr_to:
                            logger.warning(f"Banned Qbitcoin Address found in transfer_token.addr_to at block {current_block_number}")
                            return False
                elif tx_type == 'slave':
                    for slave_pk in self.pbdata.slave.slave_pks:
                        slave_addr = falcon_pk_to_address(slave_pk)
                        if banned_address == slave_addr:
                            logger.warning(f"Banned Qbitcoin Address found in slave addresses at block {current_block_number}")
                            return False
                elif tx_type == 'latticePK':
                    # Lattice transactions don't have recipient addresses
                    # Only sender validation (already done above with addr_from_pk)
                    pass
                elif tx_type == 'multi_sig_create':
                    # Check all signatories in multi-sig creation
                    for signatory in self.pbdata.multi_sig_create.signatories:
                        if banned_address == signatory:
                            logger.warning(f"Banned Qbitcoin Address found in multi_sig_create.signatories at block {current_block_number}")
                            return False
                elif tx_type == 'multi_sig_spend':
                    # Check all addresses to in multi-sig spend
                    for addr_to in self.pbdata.multi_sig_spend.addrs_to:
                        if banned_address == addr_to:
                            logger.warning(f"Banned Qbitcoin Address found in multi_sig_spend.addr_to at block {current_block_number}")
                            return False
                elif tx_type == 'multi_sig_vote':
                    # Multi-sig vote transactions don't have recipient addresses
                    # Only sender validation (already done above with addr_from_pk)
                    pass

        if self.pbdata.WhichOneof('transactionType') == 'coinbase':
            if not self._validate_extended(state_container):
                return False
            return True

        if not self.validate(True):  # It also calls _validate_custom
            return False
        if not self.validate_slave(state_container):
            return False
        if not self._validate_extended(state_container):
            return False

        addr_from_pk = falcon_pk_to_address(self.PK)
        addr_from_pk_state = state_container.addresses_state[addr_from_pk]

        expected_nonce = addr_from_pk_state.nonce + 1

        if check_nonce and self.nonce != expected_nonce:
            logger.warning('nonce incorrect, invalid tx')
            logger.warning('subtype: %s', self.type)
            logger.warning('%s actual: %s expected: %s',
                           OptimizedAddressState.bin_to_qaddress(addr_from_pk),
                           self.nonce,
                           expected_nonce)
            return False

        # Falcon public key reuse validation removed - allowing multiple transactions with same public key

        return True

    # TODO: will need state_container
    def _coinbase_filter(self):
        if config.dev.coinbase_address in [falcon_pk_to_address(self.PK), self.master_addr]:
            raise ValueError('Coinbase Address only allowed to do Coinbase Transaction')

    def _get_allowed_access_types(self):
        return [0]

    def _get_master_address(self):
        return self.addr_from

    # TODO: will need state_container
    def validate_or_raise(self, verify_signature=True) -> bool:
        """
        This method will validate a transaction and raise exception if problems are found
        :return: True if the exception is valid, exceptions otherwise
        :rtype: bool
        """
        if not self._validate_custom():
            raise ValueError("Custom validation failed")

        self._coinbase_filter()

        expected_transaction_hash = self.generate_txhash()
        if verify_signature and self.txhash != expected_transaction_hash:
            logger.warning('Invalid Transaction hash')
            logger.warning('Expected Transaction hash %s', bin2hstr(expected_transaction_hash))
            logger.warning('Found Transaction hash %s', bin2hstr(self.txhash))
            raise ValueError("Invalid Transaction Hash")

        if verify_signature:
            # Use Falcon512 for signature verification
            if not FalconSignature.verify_signature(self.get_data_hash(),
                                                    self.signature,
                                                    self.PK):
                raise ValueError("Invalid Falcon512 signature")

        return True

    def validate_slave(self, state_container: StateContainer) -> bool:
        addr_from_pk = falcon_pk_to_address(self.PK)

        master_address = self._get_master_address()
        allowed_access_types = self._get_allowed_access_types()

        # If master_addr is not set or equals addr_from_pk, this is a normal transaction
        if not self.master_addr or self.master_addr == addr_from_pk:
            return True

        # This is a slave transaction - master_addr is set and different from addr_from_pk
        if addr_from_pk != master_address:
            if (self.addr_from, self.PK) not in state_container.slaves.data:
                logger.warning("Public key and address doesn't match")
                return False

            slave_access_type = state_container.slaves.data[(self.addr_from, self.PK)].access_type
            if slave_access_type not in allowed_access_types:
                logger.warning('Access Type %s', slave_access_type)
                logger.warning('Slave Address doesnt have sufficient permission')
                return False

        return True

    def get_message_hash(self):
        # FIXME: refactor, review that things are not recalculated too often, cache, etc.
        return self.txhash

    def to_json(self):
        # FIXME: Remove once we move completely to protobuf
        return MessageToJson(self._data, sort_keys=True)

    def serialize(self) -> str:
        return self._data.SerializeToString()

    @staticmethod
    def deserialize(data):
        pbdata = qbit_pb2.Transaction()
        pbdata.ParseFromString(bytes(data))
        tx = Transaction(pbdata)
        return tx

    def _apply_state_changes_for_PK(self,
                                    state_container: StateContainer) -> bool:
        addr_from_pk = falcon_pk_to_address(self.PK)
        address_state = state_container.addresses_state[addr_from_pk]
        if self.addr_from != addr_from_pk:
            state_container.paginated_tx_hash.insert(address_state, self.txhash)
        address_state.increase_nonce()
        # Track the Falcon public key instead of OTS key
        state_container.paginated_bitfield.set_ots_key(state_container.addresses_state, addr_from_pk, self.PK)

        return True

    def _revert_state_changes_for_PK(self,
                                     state_container: StateContainer) -> bool:
        addr_from_pk = falcon_pk_to_address(self.PK)
        address_state = state_container.addresses_state[addr_from_pk]
        if self.addr_from != addr_from_pk:
            state_container.paginated_tx_hash.remove(address_state, self.txhash)
        address_state.decrease_nonce()
        # Remove the Falcon public key instead of OTS key
        state_container.paginated_bitfield.unset_ots_key(state_container.addresses_state, addr_from_pk, self.PK)

        return True



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: qbitwallet.proto
# Protobuf Python Version: 6.30.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    30,
    0,
    '',
    'qbitwallet.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


import qbitcoin.generated.qbit_pb2 as qbit__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x10qbitwallet.proto\x12\x03qrl\x1a\nqbit.proto\"9\n\x10\x41\x64\x64NewAddressReq\x12\x0e\n\x06height\x18\x01 \x01(\x04\x12\x15\n\rhash_function\x18\x02 \x01(\t\"A\n\x11\x41\x64\x64NewAddressResp\x12\x0c\n\x04\x63ode\x18\x01 \x01(\r\x12\r\n\x05\x65rror\x18\x02 \x01(\t\x12\x0f\n\x07\x61\x64\x64ress\x18\x03 \x01(\t\"]\n\x1a\x41\x64\x64NewAddressWithSlavesReq\x12\x0e\n\x06height\x18\x01 \x01(\x04\x12\x18\n\x10number_of_slaves\x18\x02 \x01(\x04\x12\x15\n\rhash_function\x18\x03 \x01(\t\"%\n\x15\x41\x64\x64\x41\x64\x64ressFromSeedReq\x12\x0c\n\x04seed\x18\x01 \x01(\t\"F\n\x16\x41\x64\x64\x41\x64\x64ressFromSeedResp\x12\x0c\n\x04\x63ode\x18\x01 \x01(\r\x12\r\n\x05\x65rror\x18\x02 \x01(\t\x12\x0f\n\x07\x61\x64\x64ress\x18\x03 \x01(\t\"\x12\n\x10ListAddressesReq\"C\n\x11ListAddressesResp\x12\x0c\n\x04\x63ode\x18\x01 \x01(\r\x12\r\n\x05\x65rror\x18\x02 \x01(\t\x12\x11\n\taddresses\x18\x03 \x03(\t\"#\n\x10RemoveAddressReq\x12\x0f\n\x07\x61\x64\x64ress\x18\x01 \x01(\t\"0\n\x11RemoveAddressResp\x12\x0c\n\x04\x63ode\x18\x01 \x01(\r\x12\r\n\x05\x65rror\x18\x02 \x01(\t\"\"\n\x0fValidAddressReq\x12\x0f\n\x07\x61\x64\x64ress\x18\x01 \x01(\t\">\n\x10ValidAddressResp\x12\x0c\n\x04\x63ode\x18\x01 \x01(\r\x12\r\n\x05\x65rror\x18\x02 \x01(\t\x12\r\n\x05valid\x18\x03 \x01(\t\"&\n\x10\x45ncryptWalletReq\x12\x12\n\npassphrase\x18\x01 \x01(\t\"0\n\x11\x45ncryptWalletResp\x12\x0c\n\x04\x63ode\x18\x01 \x01(\r\x12\r\n\x05\x65rror\x18\x02 \x01(\t\"%\n\x0fUnlockWalletReq\x12\x12\n\npassphrase\x18\x01 \x01(\t\"/\n\x10UnlockWalletResp\x12\x0c\n\x04\x63ode\x18\x01 \x01(\r\x12\r\n\x05\x65rror\x18\x02 \x01(\t\"\x0f\n\rLockWalletReq\"-\n\x0eLockWalletResp\x12\x0c\n\x04\x63ode\x18\x01 \x01(\r\x12\r\n\x05\x65rror\x18\x02 \x01(\t\"&\n\x13GetRecoverySeedsReq\x12\x0f\n\x07\x61\x64\x64ress\x18\x01 \x01(\t\"V\n\x14GetRecoverySeedsResp\x12\x0c\n\x04\x63ode\x18\x01 \x01(\r\x12\r\n\x05\x65rror\x18\x02 \x01(\t\x12\x0f\n\x07hexseed\x18\x03 \x01(\t\x12\x10\n\x08mnemonic\x18\x04 \x01(\t\"\x12\n\x10GetWalletInfoReq\"n\n\x11GetWalletInfoResp\x12\x0c\n\x04\x63ode\x18\x01 \x01(\r\x12\r\n\x05\x65rror\x18\x02 \x01(\t\x12\x0f\n\x07version\x18\x03 \x01(\r\x12\x15\n\raddress_count\x18\x04 \x01(\x04\x12\x14\n\x0cis_encrypted\x18\x05 \x01(\x08\"\x8c\x01\n\x13RelayTransferTxnReq\x12\x14\n\x0c\x61\x64\x64resses_to\x18\x01 \x03(\t\x12\x0f\n\x07\x61mounts\x18\x02 \x03(\x04\x12\x0b\n\x03\x66\x65\x65\x18\x03 \x01(\x04\x12\x16\n\x0emaster_address\x18\x04 \x01(\t\x12\x16\n\x0esigner_address\x18\x05 \x01(\t\x12\x11\n\tots_index\x18\x06 \x01(\x04\"h\n\x1aRelayTransferTxnBySlaveReq\x12\x14\n\x0c\x61\x64\x64resses_to\x18\x01 \x03(\t\x12\x0f\n\x07\x61mounts\x18\x02 \x03(\x04\x12\x0b\n\x03\x66\x65\x65\x18\x03 \x01(\x04\x12\x16\n\x0emaster_address\x18\x04 \x01(\t\"u\n\x12RelayMessageTxnReq\x12\x0f\n\x07message\x18\x01 \x01(\t\x12\x0b\n\x03\x66\x65\x65\x18\x02 \x01(\x04\x12\x16\n\x0emaster_address\x18\x03 \x01(\t\x12\x16\n\x0esigner_address\x18\x04 \x01(\t\x12\x11\n\tots_index\x18\x05 \x01(\x04\"Q\n\x19RelayMessageTxnBySlaveReq\x12\x0f\n\x07message\x18\x01 \x01(\t\x12\x0b\n\x03\x66\x65\x65\x18\x02 \x01(\x04\x12\x16\n\x0emaster_address\x18\x03 \x01(\t\"\xc5\x01\n\x10RelayTokenTxnReq\x12\x0e\n\x06symbol\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\r\n\x05owner\x18\x03 \x01(\t\x12\x10\n\x08\x64\x65\x63imals\x18\x04 \x01(\x04\x12\x11\n\taddresses\x18\x05 \x03(\t\x12\x0f\n\x07\x61mounts\x18\x06 \x03(\x04\x12\x0b\n\x03\x66\x65\x65\x18\x07 \x01(\x04\x12\x16\n\x0emaster_address\x18\x08 \x01(\t\x12\x16\n\x0esigner_address\x18\t \x01(\t\x12\x11\n\tots_index\x18\n \x01(\x04\"\xa1\x01\n\x17RelayTokenTxnBySlaveReq\x12\x0e\n\x06symbol\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\r\n\x05owner\x18\x03 \x01(\t\x12\x10\n\x08\x64\x65\x63imals\x18\x04 \x01(\x04\x12\x11\n\taddresses\x18\x05 \x03(\t\x12\x0f\n\x07\x61mounts\x18\x06 \x03(\x04\x12\x0b\n\x03\x66\x65\x65\x18\x07 \x01(\x04\x12\x16\n\x0emaster_address\x18\x08 \x01(\t\"\xa7\x01\n\x18RelayTransferTokenTxnReq\x12\x14\n\x0c\x61\x64\x64resses_to\x18\x01 \x03(\t\x12\x14\n\x0ctoken_txhash\x18\x02 \x01(\t\x12\x0f\n\x07\x61mounts\x18\x03 \x03(\x04\x12\x0b\n\x03\x66\x65\x65\x18\x04 \x01(\x04\x12\x16\n\x0emaster_address\x18\x05 \x01(\t\x12\x16\n\x0esigner_address\x18\x06 \x01(\t\x12\x11\n\tots_index\x18\x07 \x01(\x04\"\x83\x01\n\x1fRelayTransferTokenTxnBySlaveReq\x12\x14\n\x0c\x61\x64\x64resses_to\x18\x01 \x03(\t\x12\x14\n\x0ctoken_txhash\x18\x02 \x01(\t\x12\x0f\n\x07\x61mounts\x18\x03 \x03(\x04\x12\x0b\n\x03\x66\x65\x65\x18\x04 \x01(\x04\x12\x16\n\x0emaster_address\x18\x05 \x01(\t\"\x8b\x01\n\x10RelaySlaveTxnReq\x12\x11\n\tslave_pks\x18\x01 \x03(\x0c\x12\x14\n\x0c\x61\x63\x63\x65ss_types\x18\x02 \x03(\r\x12\x0b\n\x03\x66\x65\x65\x18\x03 \x01(\x04\x12\x16\n\x0emaster_address\x18\x04 \x01(\t\x12\x16\n\x0esigner_address\x18\x05 \x01(\t\x12\x11\n\tots_index\x18\x06 \x01(\x04\"g\n\x17RelaySlaveTxnBySlaveReq\x12\x11\n\tslave_pks\x18\x01 \x03(\x0c\x12\x14\n\x0c\x61\x63\x63\x65ss_types\x18\x02 \x03(\r\x12\x0b\n\x03\x66\x65\x65\x18\x03 \x01(\x04\x12\x16\n\x0emaster_address\x18\x04 \x01(\t\"N\n\x0cRelayTxnResp\x12\x0c\n\x04\x63ode\x18\x01 \x01(\r\x12\r\n\x05\x65rror\x18\x02 \x01(\t\x12!\n\x02tx\x18\x03 \x01(\x0b\x32\x15.qrl.PlainTransaction\"C\n\x13\x43hangePassphraseReq\x12\x15\n\roldPassphrase\x18\x01 \x01(\t\x12\x15\n\rnewPassphrase\x18\x02 \x01(\t\"3\n\x14\x43hangePassphraseResp\x12\x0c\n\x04\x63ode\x18\x01 \x01(\r\x12\r\n\x05\x65rror\x18\x02 \x01(\t\"+\n\x18TransactionsByAddressReq\x12\x0f\n\x07\x61\x64\x64ress\x18\x01 \x01(\t\"z\n\x19TransactionsByAddressResp\x12\x0c\n\x04\x63ode\x18\x01 \x01(\r\x12\r\n\x05\x65rror\x18\x02 \x01(\t\x12/\n\x11mini_transactions\x18\x03 \x03(\x0b\x32\x14.qrl.MiniTransaction\x12\x0f\n\x07\x62\x61lance\x18\x04 \x01(\x04\"`\n!PaginatedTransactionsByAddressReq\x12\x0f\n\x07\x61\x64\x64ress\x18\x01 \x01(\t\x12\x15\n\ritem_per_page\x18\x02 \x01(\x04\x12\x13\n\x0bpage_number\x18\x03 \x01(\x04\"\x83\x01\n\"PaginatedTransactionsByAddressResp\x12\x0c\n\x04\x63ode\x18\x01 \x01(\r\x12\r\n\x05\x65rror\x18\x02 \x01(\t\x12/\n\x11mini_transactions\x18\x03 \x03(\x0b\x32\x14.qrl.MiniTransaction\x12\x0f\n\x07\x62\x61lance\x18\x04 \x01(\x04\"!\n\x0eTransactionReq\x12\x0f\n\x07tx_hash\x18\x01 \x01(\t\"\x99\x01\n\x0fTransactionResp\x12\x0c\n\x04\x63ode\x18\x01 \x01(\r\x12\r\n\x05\x65rror\x18\x02 \x01(\t\x12!\n\x02tx\x18\x03 \x01(\x0b\x32\x15.qrl.PlainTransaction\x12\x15\n\rconfirmations\x18\x04 \x01(\t\x12\x14\n\x0c\x62lock_number\x18\x05 \x01(\x04\x12\x19\n\x11\x62lock_header_hash\x18\x06 \x01(\t\"\x1d\n\nBalanceReq\x12\x0f\n\x07\x61\x64\x64ress\x18\x01 \x01(\t\";\n\x0b\x42\x61lanceResp\x12\x0c\n\x04\x63ode\x18\x01 \x01(\r\x12\r\n\x05\x65rror\x18\x02 \x01(\t\x12\x0f\n\x07\x62\x61lance\x18\x03 \x01(\t\"\x11\n\x0fTotalBalanceReq\"@\n\x10TotalBalanceResp\x12\x0c\n\x04\x63ode\x18\x01 \x01(\r\x12\r\n\x05\x65rror\x18\x02 \x01(\t\x12\x0f\n\x07\x62\x61lance\x18\x03 \x01(\t\"\x19\n\x06OTSReq\x12\x0f\n\x07\x61\x64\x64ress\x18\x01 \x01(\t\"\x9b\x01\n\x07OTSResp\x12\x0c\n\x04\x63ode\x18\x01 \x01(\r\x12\r\n\x05\x65rror\x18\x02 \x01(\t\x12\x34\n\x14ots_bitfield_by_page\x18\x03 \x03(\x0b\x32\x16.qrl.OTSBitfieldByPage\x12\x1d\n\x15next_unused_ots_index\x18\x04 \x01(\x04\x12\x1e\n\x16unused_ots_index_found\x18\x05 \x01(\x08\"\x0b\n\tHeightReq\"9\n\nHeightResp\x12\x0c\n\x04\x63ode\x18\x01 \x01(\r\x12\r\n\x05\x65rror\x18\x02 \x01(\t\x12\x0e\n\x06height\x18\x03 \x01(\x04\"\x1f\n\x08\x42lockReq\x12\x13\n\x0bheader_hash\x18\x01 \x01(\t\"H\n\tBlockResp\x12\x0c\n\x04\x63ode\x18\x01 \x01(\r\x12\r\n\x05\x65rror\x18\x02 \x01(\t\x12\x1e\n\x05\x62lock\x18\x03 \x01(\x0b\x32\x0f.qrl.PlainBlock\"(\n\x10\x42lockByNumberReq\x12\x14\n\x0c\x62lock_number\x18\x01 \x01(\x04\"\x1e\n\x10\x41\x64\x64ressFromPKReq\x12\n\n\x02pk\x18\x01 \x01(\t\"A\n\x11\x41\x64\x64ressFromPKResp\x12\x0c\n\x04\x63ode\x18\x01 \x01(\r\x12\r\n\x05\x65rror\x18\x02 \x01(\t\x12\x0f\n\x07\x61\x64\x64ress\x18\x03 \x01(\t\"\r\n\x0bNodeInfoReq\"\xc1\x01\n\x0cNodeInfoResp\x12\x0c\n\x04\x63ode\x18\x01 \x01(\r\x12\r\n\x05\x65rror\x18\x02 \x01(\t\x12\x0f\n\x07version\x18\x03 \x01(\t\x12\x17\n\x0fnum_connections\x18\x04 \x01(\t\x12\x17\n\x0fnum_known_peers\x18\x05 \x01(\t\x12\x0e\n\x06uptime\x18\x06 \x01(\x04\x12\x14\n\x0c\x62lock_height\x18\x07 \x01(\x04\x12\x17\n\x0f\x62lock_last_hash\x18\x08 \x01(\t\x12\x12\n\nnetwork_id\x18\t \x01(\t\"7\n\x13PlainGenesisBalance\x12\x0f\n\x07\x61\x64\x64ress\x18\x01 \x01(\t\x12\x0f\n\x07\x62\x61lance\x18\x02 \x01(\x04\"5\n\x12PlainAddressAmount\x12\x0f\n\x07\x61\x64\x64ress\x18\x01 \x01(\t\x12\x0e\n\x06\x61mount\x18\x02 \x01(\x04\"\xce\x07\n\x10PlainTransaction\x12\x13\n\x0bmaster_addr\x18\x01 \x01(\t\x12\x0b\n\x03\x66\x65\x65\x18\x02 \x01(\x04\x12\x12\n\npublic_key\x18\x03 \x01(\t\x12\x11\n\tsignature\x18\x04 \x01(\t\x12\r\n\x05nonce\x18\x05 \x01(\x04\x12\x18\n\x10transaction_hash\x18\x06 \x01(\t\x12\x13\n\x0bsigner_addr\x18\x07 \x01(\t\x12\x32\n\x08transfer\x18\x08 \x01(\x0b\x32\x1e.qrl.PlainTransaction.TransferH\x00\x12\x32\n\x08\x63oinbase\x18\t \x01(\x0b\x32\x1e.qrl.PlainTransaction.CoinBaseH\x00\x12;\n\tlatticePK\x18\n \x01(\x0b\x32&.qrl.PlainTransaction.LatticePublicKeyH\x00\x12\x30\n\x07message\x18\x0b \x01(\x0b\x32\x1d.qrl.PlainTransaction.MessageH\x00\x12,\n\x05token\x18\x0c \x01(\x0b\x32\x1b.qrl.PlainTransaction.TokenH\x00\x12=\n\x0etransfer_token\x18\r \x01(\x0b\x32#.qrl.PlainTransaction.TransferTokenH\x00\x12,\n\x05slave\x18\x0e \x01(\x0b\x32\x1b.qrl.PlainTransaction.SlaveH\x00\x1a-\n\x08Transfer\x12\x10\n\x08\x61\x64\x64rs_to\x18\x01 \x03(\t\x12\x0f\n\x07\x61mounts\x18\x02 \x03(\x04\x1a+\n\x08\x43oinBase\x12\x0f\n\x07\x61\x64\x64r_to\x18\x01 \x01(\t\x12\x0e\n\x06\x61mount\x18\x02 \x01(\x04\x1a:\n\x10LatticePublicKey\x12\x10\n\x08kyber_pk\x18\x01 \x01(\t\x12\x14\n\x0c\x64ilithium_pk\x18\x02 \x01(\t\x1a\x1f\n\x07Message\x12\x14\n\x0cmessage_hash\x18\x01 \x01(\t\x1ay\n\x05Token\x12\x0e\n\x06symbol\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\r\n\x05owner\x18\x03 \x01(\t\x12\x10\n\x08\x64\x65\x63imals\x18\x04 \x01(\x04\x12\x31\n\x10initial_balances\x18\x05 \x03(\x0b\x32\x17.qrl.PlainAddressAmount\x1aH\n\rTransferToken\x12\x14\n\x0ctoken_txhash\x18\x01 \x01(\t\x12\x10\n\x08\x61\x64\x64rs_to\x18\x02 \x03(\t\x12\x0f\n\x07\x61mounts\x18\x03 \x03(\x04\x1a\x30\n\x05Slave\x12\x11\n\tslave_pks\x18\x01 \x03(\t\x12\x14\n\x0c\x61\x63\x63\x65ss_types\x18\x02 \x03(\rB\x11\n\x0ftransactionType\"\xdc\x01\n\x10PlainBlockHeader\x12\x13\n\x0bhash_header\x18\x01 \x01(\t\x12\x14\n\x0c\x62lock_number\x18\x02 \x01(\x04\x12\x19\n\x11timestamp_seconds\x18\x03 \x01(\x04\x12\x18\n\x10hash_header_prev\x18\x04 \x01(\t\x12\x14\n\x0creward_block\x18\x05 \x01(\x04\x12\x12\n\nreward_fee\x18\x06 \x01(\x04\x12\x13\n\x0bmerkle_root\x18\x07 \x01(\t\x12\x14\n\x0cmining_nonce\x18\x08 \x01(\r\x12\x13\n\x0b\x65xtra_nonce\x18\t \x01(\x04\"\x93\x01\n\nPlainBlock\x12%\n\x06header\x18\x01 \x01(\x0b\x32\x15.qrl.PlainBlockHeader\x12+\n\x0ctransactions\x18\x02 \x03(\x0b\x32\x15.qrl.PlainTransaction\x12\x31\n\x0fgenesis_balance\x18\x03 \x03(\x0b\x32\x18.qrl.PlainGenesisBalance2\xdb\x10\n\tWalletAPI\x12>\n\rAddNewAddress\x12\x15.qrl.AddNewAddressReq\x1a\x16.qrl.AddNewAddressResp\x12R\n\x17\x41\x64\x64NewAddressWithSlaves\x12\x1f.qrl.AddNewAddressWithSlavesReq\x1a\x16.qrl.AddNewAddressResp\x12>\n\rListAddresses\x12\x15.qrl.ListAddressesReq\x1a\x16.qrl.ListAddressesResp\x12>\n\rRemoveAddress\x12\x15.qrl.RemoveAddressReq\x1a\x16.qrl.RemoveAddressResp\x12=\n\x0eIsValidAddress\x12\x14.qrl.ValidAddressReq\x1a\x15.qrl.ValidAddressResp\x12>\n\rEncryptWallet\x12\x15.qrl.EncryptWalletReq\x1a\x16.qrl.EncryptWalletResp\x12\x35\n\nLockWallet\x12\x12.qrl.LockWalletReq\x1a\x13.qrl.LockWalletResp\x12;\n\x0cUnlockWallet\x12\x14.qrl.UnlockWalletReq\x1a\x15.qrl.UnlockWalletResp\x12G\n\x10GetRecoverySeeds\x12\x18.qrl.GetRecoverySeedsReq\x1a\x19.qrl.GetRecoverySeedsResp\x12>\n\rGetWalletInfo\x12\x15.qrl.GetWalletInfoReq\x1a\x16.qrl.GetWalletInfoResp\x12?\n\x10RelayTransferTxn\x12\x18.qrl.RelayTransferTxnReq\x1a\x11.qrl.RelayTxnResp\x12M\n\x17RelayTransferTxnBySlave\x12\x1f.qrl.RelayTransferTxnBySlaveReq\x1a\x11.qrl.RelayTxnResp\x12=\n\x0fRelayMessageTxn\x12\x17.qrl.RelayMessageTxnReq\x1a\x11.qrl.RelayTxnResp\x12K\n\x16RelayMessageTxnBySlave\x12\x1e.qrl.RelayMessageTxnBySlaveReq\x1a\x11.qrl.RelayTxnResp\x12\x39\n\rRelayTokenTxn\x12\x15.qrl.RelayTokenTxnReq\x1a\x11.qrl.RelayTxnResp\x12G\n\x14RelayTokenTxnBySlave\x12\x1c.qrl.RelayTokenTxnBySlaveReq\x1a\x11.qrl.RelayTxnResp\x12I\n\x15RelayTransferTokenTxn\x12\x1d.qrl.RelayTransferTokenTxnReq\x1a\x11.qrl.RelayTxnResp\x12W\n\x1cRelayTransferTokenTxnBySlave\x12$.qrl.RelayTransferTokenTxnBySlaveReq\x1a\x11.qrl.RelayTxnResp\x12\x39\n\rRelaySlaveTxn\x12\x15.qrl.RelaySlaveTxnReq\x1a\x11.qrl.RelayTxnResp\x12G\n\x14RelaySlaveTxnBySlave\x12\x1c.qrl.RelaySlaveTxnBySlaveReq\x1a\x11.qrl.RelayTxnResp\x12G\n\x10\x43hangePassphrase\x12\x18.qrl.ChangePassphraseReq\x1a\x19.qrl.ChangePassphraseResp\x12Y\n\x18GetTransactionsByAddress\x12\x1d.qrl.TransactionsByAddressReq\x1a\x1e.qrl.TransactionsByAddressResp\x12t\n!GetPaginatedTransactionsByAddress\x12&.qrl.PaginatedTransactionsByAddressReq\x1a\'.qrl.PaginatedTransactionsByAddressResp\x12;\n\x0eGetTransaction\x12\x13.qrl.TransactionReq\x1a\x14.qrl.TransactionResp\x12/\n\nGetBalance\x12\x0f.qrl.BalanceReq\x1a\x10.qrl.BalanceResp\x12>\n\x0fGetTotalBalance\x12\x14.qrl.TotalBalanceReq\x1a\x15.qrl.TotalBalanceResp\x12#\n\x06GetOTS\x12\x0b.qrl.OTSReq\x1a\x0c.qrl.OTSResp\x12,\n\tGetHeight\x12\x0e.qrl.HeightReq\x1a\x0f.qrl.HeightResp\x12)\n\x08GetBlock\x12\r.qrl.BlockReq\x1a\x0e.qrl.BlockResp\x12\x39\n\x10GetBlockByNumber\x12\x15.qrl.BlockByNumberReq\x1a\x0e.qrl.BlockResp\x12\x41\n\x10GetAddressFromPK\x12\x15.qrl.AddressFromPKReq\x1a\x16.qrl.AddressFromPKResp\x12\x32\n\x0bGetNodeInfo\x12\x10.qrl.NodeInfoReq\x1a\x11.qrl.NodeInfoRespb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'qbitwallet_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_ADDNEWADDRESSREQ']._serialized_start=37
  _globals['_ADDNEWADDRESSREQ']._serialized_end=94
  _globals['_ADDNEWADDRESSRESP']._serialized_start=96
  _globals['_ADDNEWADDRESSRESP']._serialized_end=161
  _globals['_ADDNEWADDRESSWITHSLAVESREQ']._serialized_start=163
  _globals['_ADDNEWADDRESSWITHSLAVESREQ']._serialized_end=256
  _globals['_ADDADDRESSFROMSEEDREQ']._serialized_start=258
  _globals['_ADDADDRESSFROMSEEDREQ']._serialized_end=295
  _globals['_ADDADDRESSFROMSEEDRESP']._serialized_start=297
  _globals['_ADDADDRESSFROMSEEDRESP']._serialized_end=367
  _globals['_LISTADDRESSESREQ']._serialized_start=369
  _globals['_LISTADDRESSESREQ']._serialized_end=387
  _globals['_LISTADDRESSESRESP']._serialized_start=389
  _globals['_LISTADDRESSESRESP']._serialized_end=456
  _globals['_REMOVEADDRESSREQ']._serialized_start=458
  _globals['_REMOVEADDRESSREQ']._serialized_end=493
  _globals['_REMOVEADDRESSRESP']._serialized_start=495
  _globals['_REMOVEADDRESSRESP']._serialized_end=543
  _globals['_VALIDADDRESSREQ']._serialized_start=545
  _globals['_VALIDADDRESSREQ']._serialized_end=579
  _globals['_VALIDADDRESSRESP']._serialized_start=581
  _globals['_VALIDADDRESSRESP']._serialized_end=643
  _globals['_ENCRYPTWALLETREQ']._serialized_start=645
  _globals['_ENCRYPTWALLETREQ']._serialized_end=683
  _globals['_ENCRYPTWALLETRESP']._serialized_start=685
  _globals['_ENCRYPTWALLETRESP']._serialized_end=733
  _globals['_UNLOCKWALLETREQ']._serialized_start=735
  _globals['_UNLOCKWALLETREQ']._serialized_end=772
  _globals['_UNLOCKWALLETRESP']._serialized_start=774
  _globals['_UNLOCKWALLETRESP']._serialized_end=821
  _globals['_LOCKWALLETREQ']._serialized_start=823
  _globals['_LOCKWALLETREQ']._serialized_end=838
  _globals['_LOCKWALLETRESP']._serialized_start=840
  _globals['_LOCKWALLETRESP']._serialized_end=885
  _globals['_GETRECOVERYSEEDSREQ']._serialized_start=887
  _globals['_GETRECOVERYSEEDSREQ']._serialized_end=925
  _globals['_GETRECOVERYSEEDSRESP']._serialized_start=927
  _globals['_GETRECOVERYSEEDSRESP']._serialized_end=1013
  _globals['_GETWALLETINFOREQ']._serialized_start=1015
  _globals['_GETWALLETINFOREQ']._serialized_end=1033
  _globals['_GETWALLETINFORESP']._serialized_start=1035
  _globals['_GETWALLETINFORESP']._serialized_end=1145
  _globals['_RELAYTRANSFERTXNREQ']._serialized_start=1148
  _globals['_RELAYTRANSFERTXNREQ']._serialized_end=1288
  _globals['_RELAYTRANSFERTXNBYSLAVEREQ']._serialized_start=1290
  _globals['_RELAYTRANSFERTXNBYSLAVEREQ']._serialized_end=1394
  _globals['_RELAYMESSAGETXNREQ']._serialized_start=1396
  _globals['_RELAYMESSAGETXNREQ']._serialized_end=1513
  _globals['_RELAYMESSAGETXNBYSLAVEREQ']._serialized_start=1515
  _globals['_RELAYMESSAGETXNBYSLAVEREQ']._serialized_end=1596
  _globals['_RELAYTOKENTXNREQ']._serialized_start=1599
  _globals['_RELAYTOKENTXNREQ']._serialized_end=1796
  _globals['_RELAYTOKENTXNBYSLAVEREQ']._serialized_start=1799
  _globals['_RELAYTOKENTXNBYSLAVEREQ']._serialized_end=1960
  _globals['_RELAYTRANSFERTOKENTXNREQ']._serialized_start=1963
  _globals['_RELAYTRANSFERTOKENTXNREQ']._serialized_end=2130
  _globals['_RELAYTRANSFERTOKENTXNBYSLAVEREQ']._serialized_start=2133
  _globals['_RELAYTRANSFERTOKENTXNBYSLAVEREQ']._serialized_end=2264
  _globals['_RELAYSLAVETXNREQ']._serialized_start=2267
  _globals['_RELAYSLAVETXNREQ']._serialized_end=2406
  _globals['_RELAYSLAVETXNBYSLAVEREQ']._serialized_start=2408
  _globals['_RELAYSLAVETXNBYSLAVEREQ']._serialized_end=2511
  _globals['_RELAYTXNRESP']._serialized_start=2513
  _globals['_RELAYTXNRESP']._serialized_end=2591
  _globals['_CHANGEPASSPHRASEREQ']._serialized_start=2593
  _globals['_CHANGEPASSPHRASEREQ']._serialized_end=2660
  _globals['_CHANGEPASSPHRASERESP']._serialized_start=2662
  _globals['_CHANGEPASSPHRASERESP']._serialized_end=2713
  _globals['_TRANSACTIONSBYADDRESSREQ']._serialized_start=2715
  _globals['_TRANSACTIONSBYADDRESSREQ']._serialized_end=2758
  _globals['_TRANSACTIONSBYADDRESSRESP']._serialized_start=2760
  _globals['_TRANSACTIONSBYADDRESSRESP']._serialized_end=2882
  _globals['_PAGINATEDTRANSACTIONSBYADDRESSREQ']._serialized_start=2884
  _globals['_PAGINATEDTRANSACTIONSBYADDRESSREQ']._serialized_end=2980
  _globals['_PAGINATEDTRANSACTIONSBYADDRESSRESP']._serialized_start=2983
  _globals['_PAGINATEDTRANSACTIONSBYADDRESSRESP']._serialized_end=3114
  _globals['_TRANSACTIONREQ']._serialized_start=3116
  _globals['_TRANSACTIONREQ']._serialized_end=3149
  _globals['_TRANSACTIONRESP']._serialized_start=3152
  _globals['_TRANSACTIONRESP']._serialized_end=3305
  _globals['_BALANCEREQ']._serialized_start=3307
  _globals['_BALANCEREQ']._serialized_end=3336
  _globals['_BALANCERESP']._serialized_start=3338
  _globals['_BALANCERESP']._serialized_end=3397
  _globals['_TOTALBALANCEREQ']._serialized_start=3399
  _globals['_TOTALBALANCEREQ']._serialized_end=3416
  _globals['_TOTALBALANCERESP']._serialized_start=3418
  _globals['_TOTALBALANCERESP']._serialized_end=3482
  _globals['_OTSREQ']._serialized_start=3484
  _globals['_OTSREQ']._serialized_end=3509
  _globals['_OTSRESP']._serialized_start=3512
  _globals['_OTSRESP']._serialized_end=3667
  _globals['_HEIGHTREQ']._serialized_start=3669
  _globals['_HEIGHTREQ']._serialized_end=3680
  _globals['_HEIGHTRESP']._serialized_start=3682
  _globals['_HEIGHTRESP']._serialized_end=3739
  _globals['_BLOCKREQ']._serialized_start=3741
  _globals['_BLOCKREQ']._serialized_end=3772
  _globals['_BLOCKRESP']._serialized_start=3774
  _globals['_BLOCKRESP']._serialized_end=3846
  _globals['_BLOCKBYNUMBERREQ']._serialized_start=3848
  _globals['_BLOCKBYNUMBERREQ']._serialized_end=3888
  _globals['_ADDRESSFROMPKREQ']._serialized_start=3890
  _globals['_ADDRESSFROMPKREQ']._serialized_end=3920
  _globals['_ADDRESSFROMPKRESP']._serialized_start=3922
  _globals['_ADDRESSFROMPKRESP']._serialized_end=3987
  _globals['_NODEINFOREQ']._serialized_start=3989
  _globals['_NODEINFOREQ']._serialized_end=4002
  _globals['_NODEINFORESP']._serialized_start=4005
  _globals['_NODEINFORESP']._serialized_end=4198
  _globals['_PLAINGENESISBALANCE']._serialized_start=4200
  _globals['_PLAINGENESISBALANCE']._serialized_end=4255
  _globals['_PLAINADDRESSAMOUNT']._serialized_start=4257
  _globals['_PLAINADDRESSAMOUNT']._serialized_end=4310
  _globals['_PLAINTRANSACTION']._serialized_start=4313
  _globals['_PLAINTRANSACTION']._serialized_end=5287
  _globals['_PLAINTRANSACTION_TRANSFER']._serialized_start=4838
  _globals['_PLAINTRANSACTION_TRANSFER']._serialized_end=4883
  _globals['_PLAINTRANSACTION_COINBASE']._serialized_start=4885
  _globals['_PLAINTRANSACTION_COINBASE']._serialized_end=4928
  _globals['_PLAINTRANSACTION_LATTICEPUBLICKEY']._serialized_start=4930
  _globals['_PLAINTRANSACTION_LATTICEPUBLICKEY']._serialized_end=4988
  _globals['_PLAINTRANSACTION_MESSAGE']._serialized_start=4990
  _globals['_PLAINTRANSACTION_MESSAGE']._serialized_end=5021
  _globals['_PLAINTRANSACTION_TOKEN']._serialized_start=5023
  _globals['_PLAINTRANSACTION_TOKEN']._serialized_end=5144
  _globals['_PLAINTRANSACTION_TRANSFERTOKEN']._serialized_start=5146
  _globals['_PLAINTRANSACTION_TRANSFERTOKEN']._serialized_end=5218
  _globals['_PLAINTRANSACTION_SLAVE']._serialized_start=5220
  _globals['_PLAINTRANSACTION_SLAVE']._serialized_end=5268
  _globals['_PLAINBLOCKHEADER']._serialized_start=5290
  _globals['_PLAINBLOCKHEADER']._serialized_end=5510
  _globals['_PLAINBLOCK']._serialized_start=5513
  _globals['_PLAINBLOCK']._serialized_end=5660
  _globals['_WALLETAPI']._serialized_start=5663
  _globals['_WALLETAPI']._serialized_end=7802
# @@protoc_insertion_point(module_scope)



from pyqrllib.pyqrllib import bin2hstr
from pyqryptonight.pyqryptonight import UInt256ToString

from qbitcoin.core.misc import logger
from qbitcoin.core.Block import Block
from qbitcoin.core.p2p.p2pObserver import P2PBaseObserver
from qbitcoin.generated import qbitlegacy_pb2, qbit_pb2


class P2PChainManager(P2PBaseObserver):
    def __init__(self):
        super().__init__()

    def new_channel(self, channel):
        channel.register(qbitlegacy_pb2.LegacyMessage.BK, self.handle_block)
        channel.register(qbitlegacy_pb2.LegacyMessage.FB, self.handle_fetch_block)
        channel.register(qbitlegacy_pb2.LegacyMessage.PB, self.handle_push_block)
        channel.register(qbitlegacy_pb2.LegacyMessage.BH, self.handle_block_height)
        channel.register(qbitlegacy_pb2.LegacyMessage.HEADERHASHES, self.handle_node_headerhash)
        # Register staker message handler using PL (peer list) message type for staker announcements
        channel.register(qbitlegacy_pb2.LegacyMessage.PL, self.handle_peer_message)

    def handle_fetch_block(self, source, message: qbitlegacy_pb2.LegacyMessage):  # Fetch Request for block
        """
        Fetch Block
        Sends the request for the block.
        :return:
        """
        P2PBaseObserver._validate_message(message, qbitlegacy_pb2.LegacyMessage.FB)

        block_number = message.fbData.index

        logger.info(' Request for %s by %s', block_number, source.peer)
        if 0 < block_number <= source.factory.chain_height:
            block = source.factory.get_block_by_number(block_number)
            msg = qbitlegacy_pb2.LegacyMessage(func_name=qbitlegacy_pb2.LegacyMessage.PB,
                                              pbData=qbitlegacy_pb2.PBData(block=block.pbdata))
            source.send(msg)

    def handle_push_block(self, source, message: qbitlegacy_pb2.LegacyMessage):
        """
        Push Block
        This function processes requested blocks received while syncing.
        Block received under this function are directly added to the main
        chain i.e. chain.blockchain
        It is expected to receive only one block for a given blocknumber.
        :return:
        """
        # FIXME: Later rename
        P2PBaseObserver._validate_message(message, qbitlegacy_pb2.LegacyMessage.PB)
        if message.pbData is None:
            return

        try:
            block = Block(message.pbData.block)
            source.factory.block_received(source, block)

        except Exception as e:
            logger.error('block rejected - unable to decode serialised data %s', source.peer)
            logger.exception(e)

    def handle_block(self, source, message: qbitlegacy_pb2.LegacyMessage):  # block received
        """
        Block
        This function processes any new block received.
        :return:
        """
        P2PBaseObserver._validate_message(message, qbitlegacy_pb2.LegacyMessage.BK)
        try:
            block = Block(message.block)
        except Exception as e:
            logger.error('block rejected - unable to decode serialised data %s', source.peer)
            logger.exception(e)
            return

        logger.info('>>>Received block from %s %s %s',
                    source.peer.full_address,
                    block.block_number,
                    bin2hstr(block.headerhash))

        if not source.factory.master_mr.isRequested(block.headerhash, source, block):
            return

        source.factory.pow.pre_block_logic(block)  # FIXME: Ignores return value
        source.factory.master_mr.register(qbitlegacy_pb2.LegacyMessage.BK, block.headerhash, message.block)

    def handle_block_height(self, source, message: qbitlegacy_pb2.LegacyMessage):
        """
        Sends / Receives Blockheight
        :param source:
        :param message:
        :return:
        """
        if message.bhData.block_number == 0:
            block = source.factory.last_block
            cumulative_difficulty = source.factory.get_cumulative_difficulty()
            if block.block_number == 0:
                return
            bhdata = qbit_pb2.BlockHeightData(block_number=block.block_number,
                                             block_headerhash=block.headerhash,
                                             cumulative_difficulty=bytes(cumulative_difficulty))
            msg = qbitlegacy_pb2.LegacyMessage(func_name=qbitlegacy_pb2.LegacyMessage.BH,
                                              bhData=bhdata)
            source.send(msg)
            return

        try:
            UInt256ToString(message.bhData.cumulative_difficulty)
        except ValueError:
            logger.warning('Invalid Block Height Data')
            source.loseConnection()
            return

        source.factory.update_peer_blockheight(source.peer.full_address,
                                               message.bhData.block_number,
                                               message.bhData.block_headerhash,
                                               message.bhData.cumulative_difficulty)

    def handle_node_headerhash(self, source, message: qbitlegacy_pb2.LegacyMessage):
        """
        Sends/Receives NodeHeaderHashes
        :param source:
        :param message:
        :return:
        """

        if len(message.nodeHeaderHash.headerhashes) == 0:
            node_headerhash = source.factory.get_headerhashes(message.nodeHeaderHash.block_number)
            msg = qbitlegacy_pb2.LegacyMessage(func_name=qbitlegacy_pb2.LegacyMessage.HEADERHASHES,
                                              nodeHeaderHash=node_headerhash)
            source.send(msg)
        else:
            source.factory.compare_and_sync(source, message.nodeHeaderHash)

    def handle_peer_message(self, source, message: qbitlegacy_pb2.LegacyMessage):
        """
        Handle peer messages which can include staker announcements
        :param source: Peer connection that sent the message
        :param message: Legacy message containing peer or staker data
        :return:
        """
        try:
            P2PBaseObserver._validate_message(message, qbitlegacy_pb2.LegacyMessage.PL)
            
            # Check if this PL message contains staker data by looking at peer_ips
            if message.plData and message.plData.peer_ips:
                for peer_ip_data in message.plData.peer_ips:
                    # Check if this is a staker announcement (starts with special prefix)
                    if peer_ip_data.startswith("STAKER:"):
                        try:
                            import json
                            # Extract staker JSON data after the "STAKER:" prefix
                            staker_json = peer_ip_data[7:]  # Remove "STAKER:" prefix
                            staker_data = json.loads(staker_json)
                            
                            # Verify this is a valid staker message
                            if staker_data.get('type') == 'staker_update':
                                # Get peer IP from the source connection 
                                peer_ip = source.transport.getPeer().host
                                
                                # Pass to staking manager if available
                                if hasattr(source.factory, '_qrl_node') and hasattr(source.factory._qrl_node, 'staking_manager'):
                                    logger.info("Received staker message from peer %s: %s %s", 
                                              peer_ip, staker_data.get('action', 'unknown'), staker_data.get('address', 'unknown'))
                                    source.factory._qrl_node.staking_manager.handle_peer_staker_message(staker_data, peer_ip)
                                else:
                                    logger.debug("No staking manager available to handle peer staker message")
                                    
                        except (json.JSONDecodeError, ValueError) as e:
                            logger.debug("Invalid staker data in peer message: %s", str(e))
                            
        except Exception as e:
            logger.error("Error handling peer message from %s: %s", source.peer, str(e))
